@@ -29,8 +29,10 @@ import pluggy
 import py
 
 import _pytest
+from _pytest._io import TerminalWriter
 from _pytest._io.saferepr import safeformat
 from _pytest._io.saferepr import saferepr
+from _pytest.compat import ATTRS_EQ_FIELD
 from _pytest.compat import overload
 from _pytest.compat import TYPE_CHECKING
 
@@ -68,19 +70,21 @@ class Code:
 
     @property
     def path(self) -> Union[py.path.local, str]:
-        """ return a path object pointing to source code (note that it
-        might not point to an actually existing file). """
+        """ return a path object pointing to source code (or a str in case
+        of OSError / non-existing file).
+        """
+        if not self.raw.co_filename:
+            return ""
         try:
             p = py.path.local(self.raw.co_filename)
             # maybe don't try this checking
             if not p.check():
                 raise OSError("py.path check failed.")
+            return p
         except OSError:
             # XXX maybe try harder like the weird logic
             # in the standard lib [linecache.updatecache] does?
-            p = self.raw.co_filename
-
-        return p
+            return self.raw.co_filename
 
     @property
     def fullsource(self) -> Optional["Source"]:
@@ -630,17 +634,18 @@ class ExceptionInfo(Generic[_E]):
         )
         return fmt.repr_excinfo(self)
 
-    def match(self, regexp: "Union[str, Pattern]") -> bool:
+    def match(self, regexp: "Union[str, Pattern]") -> "Literal[True]":
         """
-        Check whether the regular expression 'regexp' is found in the string
-        representation of the exception using ``re.search``. If it matches
-        then True is returned (so that it is possible to write
-        ``assert excinfo.match()``). If it doesn't match an AssertionError is
-        raised.
+        Check whether the regular expression `regexp` matches the string
+        representation of the exception using :func:`python:re.search`.
+        If it matches `True` is returned.
+        If it doesn't match an `AssertionError` is raised.
         """
         __tracebackhide__ = True
-        if not re.search(regexp, str(self.value)):
-            assert 0, "Pattern {!r} not found in {!r}".format(regexp, str(self.value))
+        assert re.search(
+            regexp, str(self.value)
+        ), "Pattern {!r} does not match {!r}".format(regexp, str(self.value))
+        # Return True to allow for "assert excinfo.match()".
         return True
 
 
@@ -785,11 +790,9 @@ class FormattedExcinfo:
             else:
                 message = excinfo and excinfo.typename or ""
             path = self._makepath(entry.path)
-            filelocrepr = ReprFileLocation(path, entry.lineno + 1, message)
-            localsrepr = None
-            if not short:
-                localsrepr = self.repr_locals(entry.locals)
-            return ReprEntry(lines, reprargs, localsrepr, filelocrepr, style)
+            reprfileloc = ReprFileLocation(path, entry.lineno + 1, message)
+            localsrepr = self.repr_locals(entry.locals)
+            return ReprEntry(lines, reprargs, localsrepr, reprfileloc, style)
         if excinfo:
             lines.extend(self.get_exconly(excinfo, indent=4))
         return ReprEntry(lines, None, None, None, style)
@@ -909,50 +912,53 @@ class FormattedExcinfo:
         return ExceptionChainRepr(repr_chain)
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class TerminalRepr:
     def __str__(self) -> str:
         # FYI this is called from pytest-xdist's serialization of exception
         # information.
         io = StringIO()
-        tw = py.io.TerminalWriter(file=io)
+        tw = TerminalWriter(file=io)
         self.toterminal(tw)
         return io.getvalue().strip()
 
     def __repr__(self) -> str:
         return "<{} instance at {:0x}>".format(self.__class__, id(self))
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         raise NotImplementedError()
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ExceptionRepr(TerminalRepr):
-    def __init__(self) -> None:
+    def __attrs_post_init__(self):
         self.sections = []  # type: List[Tuple[str, str, str]]
 
     def addsection(self, name: str, content: str, sep: str = "-") -> None:
         self.sections.append((name, content, sep))
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         for name, content, sep in self.sections:
             tw.sep(sep, name)
             tw.line(content)
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ExceptionChainRepr(ExceptionRepr):
-    def __init__(
-        self,
-        chain: Sequence[
+    chain = attr.ib(
+        type=Sequence[
             Tuple["ReprTraceback", Optional["ReprFileLocation"], Optional[str]]
-        ],
-    ) -> None:
-        super().__init__()
-        self.chain = chain
+        ]
+    )
+
+    def __attrs_post_init__(self):
+        super().__attrs_post_init__()
         # reprcrash and reprtraceback of the outermost (the newest) exception
         # in the chain
-        self.reprtraceback = chain[-1][0]
-        self.reprcrash = chain[-1][1]
+        self.reprtraceback = self.chain[-1][0]
+        self.reprcrash = self.chain[-1][1]
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         for element in self.chain:
             element[0].toterminal(tw)
             if element[2] is not None:
@@ -961,33 +967,25 @@ class ExceptionChainRepr(ExceptionRepr):
         super().toterminal(tw)
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprExceptionInfo(ExceptionRepr):
-    def __init__(
-        self, reprtraceback: "ReprTraceback", reprcrash: "ReprFileLocation"
-    ) -> None:
-        super().__init__()
-        self.reprtraceback = reprtraceback
-        self.reprcrash = reprcrash
+    reprtraceback = attr.ib(type="ReprTraceback")
+    reprcrash = attr.ib(type="ReprFileLocation")
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         self.reprtraceback.toterminal(tw)
         super().toterminal(tw)
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprTraceback(TerminalRepr):
+    reprentries = attr.ib(type=Sequence[Union["ReprEntry", "ReprEntryNative"]])
+    extraline = attr.ib(type=Optional[str])
+    style = attr.ib(type="_TracebackStyle")
+
     entrysep = "_ "
 
-    def __init__(
-        self,
-        reprentries: Sequence[Union["ReprEntry", "ReprEntryNative"]],
-        extraline: Optional[str],
-        style: "_TracebackStyle",
-    ) -> None:
-        self.reprentries = reprentries
-        self.extraline = extraline
-        self.style = style
-
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         # the entries might have different styles
         for i, entry in enumerate(self.reprentries):
             if entry.style == "long":
@@ -1013,44 +1011,75 @@ class ReprTracebackNative(ReprTraceback):
         self.extraline = None
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprEntryNative(TerminalRepr):
+    lines = attr.ib(type=Sequence[str])
     style = "native"  # type: _TracebackStyle
 
-    def __init__(self, tblines: Sequence[str]) -> None:
-        self.lines = tblines
-
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         tw.write("".join(self.lines))
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprEntry(TerminalRepr):
-    def __init__(
-        self,
-        lines: Sequence[str],
-        reprfuncargs: Optional["ReprFuncArgs"],
-        reprlocals: Optional["ReprLocals"],
-        filelocrepr: Optional["ReprFileLocation"],
-        style: "_TracebackStyle",
-    ) -> None:
-        self.lines = lines
-        self.reprfuncargs = reprfuncargs
-        self.reprlocals = reprlocals
-        self.reprfileloc = filelocrepr
-        self.style = style
+    lines = attr.ib(type=Sequence[str])
+    reprfuncargs = attr.ib(type=Optional["ReprFuncArgs"])
+    reprlocals = attr.ib(type=Optional["ReprLocals"])
+    reprfileloc = attr.ib(type=Optional["ReprFileLocation"])
+    style = attr.ib(type="_TracebackStyle")
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def _write_entry_lines(self, tw: TerminalWriter) -> None:
+        """Writes the source code portions of a list of traceback entries with syntax highlighting.
+
+        Usually entries are lines like these:
+
+            "     x = 1"
+            ">    assert x == 2"
+            "E    assert 1 == 2"
+
+        This function takes care of rendering the "source" portions of it (the lines without
+        the "E" prefix) using syntax highlighting, taking care to not highlighting the ">"
+        character, as doing so might break line continuations.
+        """
+
+        indent_size = 4
+
+        def is_fail(line):
+            return line.startswith("{}   ".format(FormattedExcinfo.fail_marker))
+
+        if not self.lines:
+            return
+
+        # separate indents and source lines that are not failures: we want to
+        # highlight the code but not the indentation, which may contain markers
+        # such as ">   assert 0"
+        indents = []
+        source_lines = []
+        for line in self.lines:
+            if not is_fail(line):
+                indents.append(line[:indent_size])
+                source_lines.append(line[indent_size:])
+
+        tw._write_source(source_lines, indents)
+
+        # failure lines are always completely red and bold
+        for line in (x for x in self.lines if is_fail(x)):
+            tw.line(line, bold=True, red=True)
+
+    def toterminal(self, tw: TerminalWriter) -> None:
         if self.style == "short":
             assert self.reprfileloc is not None
             self.reprfileloc.toterminal(tw)
-            for line in self.lines:
-                red = line.startswith("E   ")
-                tw.line(line, bold=True, red=red)
+            self._write_entry_lines(tw)
+            if self.reprlocals:
+                self.reprlocals.toterminal(tw, indent=" " * 8)
             return
+
         if self.reprfuncargs:
             self.reprfuncargs.toterminal(tw)
-        for line in self.lines:
-            red = line.startswith("E   ")
-            tw.line(line, bold=True, red=red)
+
+        self._write_entry_lines(tw)
+
         if self.reprlocals:
             tw.line("")
             self.reprlocals.toterminal(tw)
@@ -1065,13 +1094,13 @@ class ReprEntry(TerminalRepr):
         )
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprFileLocation(TerminalRepr):
-    def __init__(self, path, lineno: int, message: str) -> None:
-        self.path = str(path)
-        self.lineno = lineno
-        self.message = message
+    path = attr.ib(type=str, converter=str)
+    lineno = attr.ib(type=int)
+    message = attr.ib(type=str)
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         # filename and lineno output for each entry,
         # using an output format that most editors understand
         msg = self.message
@@ -1082,20 +1111,20 @@ class ReprFileLocation(TerminalRepr):
         tw.line(":{}: {}".format(self.lineno, msg))
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprLocals(TerminalRepr):
-    def __init__(self, lines: Sequence[str]) -> None:
-        self.lines = lines
+    lines = attr.ib(type=Sequence[str])
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter, indent="") -> None:
         for line in self.lines:
-            tw.line(line)
+            tw.line(indent + line)
 
 
+@attr.s(**{ATTRS_EQ_FIELD: False})  # type: ignore
 class ReprFuncArgs(TerminalRepr):
-    def __init__(self, args: Sequence[Tuple[str, object]]) -> None:
-        self.args = args
+    args = attr.ib(type=Sequence[Tuple[str, object]])
 
-    def toterminal(self, tw: py.io.TerminalWriter) -> None:
+    def toterminal(self, tw: TerminalWriter) -> None:
         if self.args:
             linesofar = ""
             for name, value in self.args:
