@@ -1,10 +1,12 @@
 from io import StringIO
 from pprint import pprint
+from typing import Any
 from typing import List
 from typing import Optional
 from typing import Tuple
 from typing import Union
 
+import attr
 import py
 
 from _pytest._code.code import ExceptionChainRepr
@@ -17,6 +19,8 @@ from _pytest._code.code import ReprFuncArgs
 from _pytest._code.code import ReprLocals
 from _pytest._code.code import ReprTraceback
 from _pytest._code.code import TerminalRepr
+from _pytest._io import TerminalWriter
+from _pytest.compat import TYPE_CHECKING
 from _pytest.nodes import Node
 from _pytest.outcomes import skip
 from _pytest.pathlib import Path
@@ -41,9 +45,14 @@ class BaseReport:
     sections = []  # type: List[Tuple[str, str]]
     nodeid = None  # type: str
 
-    def __init__(self, **kw):
+    def __init__(self, **kw: Any) -> None:
         self.__dict__.update(kw)
 
+    if TYPE_CHECKING:
+        # Can have arbitrary fields given to __init__().
+        def __getattr__(self, key: str) -> Any:
+            raise NotImplementedError()
+
     def toterminal(self, out) -> None:
         if hasattr(self, "node"):
             out.line(getslaveinfoline(self.node))  # type: ignore
@@ -73,7 +82,7 @@ class BaseReport:
 
         .. versionadded:: 3.0
         """
-        tw = py.io.TerminalWriter(stringio=True)
+        tw = TerminalWriter(stringio=True)
         tw.hasmarkup = False
         self.toterminal(tw)
         exc = tw.stringio.getvalue()
@@ -114,7 +123,7 @@ class BaseReport:
     skipped = property(lambda x: x.outcome == "skipped")
 
     @property
-    def fspath(self):
+    def fspath(self) -> str:
         return self.nodeid.split("::")[0]
 
     @property
@@ -275,8 +284,7 @@ class TestReport(BaseReport):
             if not isinstance(excinfo, ExceptionInfo):
                 outcome = "failed"
                 longrepr = excinfo
-            # Type ignored -- see comment where skip.Exception is defined.
-            elif excinfo.errisinstance(skip.Exception):  # type: ignore
+            elif excinfo.errisinstance(skip.Exception):
                 outcome = "skipped"
                 r = excinfo._getreprcrash()
                 longrepr = (str(r.path), r.lineno, r.message)
@@ -361,14 +369,14 @@ def _report_to_json(report):
     """
 
     def serialize_repr_entry(entry):
-        entry_data = {"type": type(entry).__name__, "data": entry.__dict__.copy()}
+        entry_data = {"type": type(entry).__name__, "data": attr.asdict(entry)}
         for key, value in entry_data["data"].items():
             if hasattr(value, "__dict__"):
-                entry_data["data"][key] = value.__dict__.copy()
+                entry_data["data"][key] = attr.asdict(value)
         return entry_data
 
-    def serialize_repr_traceback(reprtraceback):
-        result = reprtraceback.__dict__.copy()
+    def serialize_repr_traceback(reprtraceback: ReprTraceback):
+        result = attr.asdict(reprtraceback)
         result["reprentries"] = [
             serialize_repr_entry(x) for x in reprtraceback.reprentries
         ]
@@ -376,7 +384,7 @@ def _report_to_json(report):
 
     def serialize_repr_crash(reprcrash: Optional[ReprFileLocation]):
         if reprcrash is not None:
-            return reprcrash.__dict__.copy()
+            return attr.asdict(reprcrash)
         else:
             return None
 
@@ -443,7 +451,7 @@ def _report_kwargs_from_json(reportdict):
                 lines=data["lines"],
                 reprfuncargs=reprfuncargs,
                 reprlocals=reprlocals,
-                filelocrepr=reprfileloc,
+                reprfileloc=reprfileloc,
                 style=data["style"],
             )  # type: Union[ReprEntry, ReprEntryNative]
         elif entry_type == "ReprEntryNative":
