@@ -8,6 +8,7 @@ import datetime
 import platform
 import sys
 import time
+import warnings
 from functools import partial
 from typing import Any
 from typing import Callable
@@ -25,14 +26,29 @@ from more_itertools import collapse
 
 import pytest
 from _pytest import nodes
+from _pytest._io import TerminalWriter
 from _pytest.config import Config
-from _pytest.main import ExitCode
+from _pytest.config import ExitCode
+from _pytest.deprecated import TERMINALWRITER_WRITER
 from _pytest.main import Session
 from _pytest.reports import CollectReport
 from _pytest.reports import TestReport
 
 REPORT_COLLECTING_RESOLUTION = 0.5
 
+KNOWN_TYPES = (
+    "failed",
+    "passed",
+    "skipped",
+    "deselected",
+    "xfailed",
+    "xpassed",
+    "warnings",
+    "error",
+)
+
+_REPORTCHARS_DEFAULT = "fE"
+
 
 class MoreQuietAction(argparse.Action):
     """
@@ -68,7 +84,7 @@ def pytest_addoption(parser):
         default=0,
         dest="verbose",
         help="increase verbosity.",
-    ),
+    )
     group._addoption(
         "-q",
         "--quiet",
@@ -76,7 +92,7 @@ def pytest_addoption(parser):
         default=0,
         dest="verbose",
         help="decrease verbosity.",
-    ),
+    )
     group._addoption(
         "--verbosity",
         dest="verbose",
@@ -88,12 +104,13 @@ def pytest_addoption(parser):
         "-r",
         action="store",
         dest="reportchars",
-        default="",
+        default=_REPORTCHARS_DEFAULT,
         metavar="chars",
         help="show extra test summary info as specified by chars: (f)ailed, "
         "(E)rror, (s)kipped, (x)failed, (X)passed, "
         "(p)assed, (P)assed with output, (a)ll except passed (p/P), or (A)ll. "
-        "(w)arnings are enabled by default (see --disable-warnings).",
+        "(w)arnings are enabled by default (see --disable-warnings), "
+        "'N' can be used to reset the list. (default: 'fE').",
     )
     group._addoption(
         "--disable-warnings",
@@ -166,38 +183,42 @@ def pytest_configure(config: Config) -> None:
 
 
 def getreportopt(config: Config) -> str:
-    reportopts = ""
     reportchars = config.option.reportchars
-    if not config.option.disable_warnings and "w" not in reportchars:
-        reportchars += "w"
-    elif config.option.disable_warnings and "w" in reportchars:
-        reportchars = reportchars.replace("w", "")
+
+    old_aliases = {"F", "S"}
+    reportopts = ""
     for char in reportchars:
+        if char in old_aliases:
+            char = char.lower()
         if char == "a":
-            reportopts = "sxXwEf"
+            reportopts = "sxXEf"
         elif char == "A":
-            reportopts = "PpsxXwEf"
-            break
+            reportopts = "PpsxXEf"
+        elif char == "N":
+            reportopts = ""
         elif char not in reportopts:
             reportopts += char
+
+    if not config.option.disable_warnings and "w" not in reportopts:
+        reportopts = "w" + reportopts
+    elif config.option.disable_warnings and "w" in reportopts:
+        reportopts = reportopts.replace("w", "")
+
     return reportopts
 
 
 @pytest.hookimpl(trylast=True)  # after _pytest.runner
 def pytest_report_teststatus(report: TestReport) -> Tuple[str, str, str]:
+    letter = "F"
     if report.passed:
         letter = "."
     elif report.skipped:
         letter = "s"
-    elif report.failed:
-        letter = "F"
-        if report.when != "call":
-            letter = "f"
 
-    # Report failed CollectReports as "error" (in line with pytest_collectreport).
     outcome = report.outcome
-    if report.when == "collect" and outcome == "failed":
+    if report.when in ("collect", "setup", "teardown") and outcome == "failed":
         outcome = "error"
+        letter = "E"
 
     return outcome, letter, outcome.upper()
 
@@ -247,12 +268,12 @@ class TerminalReporter:
         self._showfspath = None
 
         self.stats = {}  # type: Dict[str, List[Any]]
+        self._main_color = None  # type: Optional[str]
+        self._known_types = None  # type: Optional[List]
         self.startdir = config.invocation_dir
         if file is None:
             file = sys.stdout
         self._tw = _pytest.config.create_terminal_writer(config, file)
-        # self.writer will be deprecated in pytest-3.4
-        self.writer = self._tw
         self._screen_width = self._tw.fullwidth
         self.currentfspath = None  # type: Any
         self.reportchars = getreportopt(config)
@@ -262,6 +283,16 @@ class TerminalReporter:
         self._show_progress_info = self._determine_show_progress_info()
         self._collect_report_last_write = None  # type: Optional[float]
 
+    @property
+    def writer(self) -> TerminalWriter:
+        warnings.warn(TERMINALWRITER_WRITER, stacklevel=2)
+        return self._tw
+
+    @writer.setter
+    def writer(self, value: TerminalWriter):
+        warnings.warn(TERMINALWRITER_WRITER, stacklevel=2)
+        self._tw = value
+
     def _determine_show_progress_info(self):
         """Return True if we should display progress information based on the current config"""
         # do not show progress if we are not capturing output (#3038)
@@ -365,6 +396,12 @@ class TerminalReporter:
     def line(self, msg, **kw):
         self._tw.line(msg, **kw)
 
+    def _add_stats(self, category: str, items: List) -> None:
+        set_main_color = category not in self.stats
+        self.stats.setdefault(category, []).extend(items[:])
+        if set_main_color:
+            self._set_main_color()
+
     def pytest_internalerror(self, excrepr):
         for line in str(excrepr).split("\n"):
             self.write_line("INTERNALERROR> " + line)
@@ -374,7 +411,6 @@ class TerminalReporter:
         # from _pytest.nodes import get_fslocation_from_item
         from _pytest.warnings import warning_record_to_str
 
-        warnings = self.stats.setdefault("warnings", [])
         fslocation = warning_message.filename, warning_message.lineno
         message = warning_record_to_str(warning_message)
 
@@ -382,7 +418,7 @@ class TerminalReporter:
         warning_report = WarningReport(
             fslocation=fslocation, message=message, nodeid=nodeid
         )
-        warnings.append(warning_report)
+        self._add_stats("warnings", [warning_report])
 
     def pytest_plugin_registered(self, plugin):
         if self.config.option.traceconfig:
@@ -393,7 +429,7 @@ class TerminalReporter:
             self.write_line(msg)
 
     def pytest_deselected(self, items):
-        self.stats.setdefault("deselected", []).extend(items)
+        self._add_stats("deselected", items)
 
     def pytest_runtest_logstart(self, nodeid, location):
         # ensure that the path is printed before the
@@ -414,7 +450,7 @@ class TerminalReporter:
             word, markup = word
         else:
             markup = None
-        self.stats.setdefault(category, []).append(rep)
+        self._add_stats(category, [rep])
         if not letter and not word:
             # probably passed setup/teardown
             return
@@ -456,6 +492,10 @@ class TerminalReporter:
                 self._tw.write(" " + line)
                 self.currentfspath = -2
 
+    @property
+    def _is_last_item(self):
+        return len(self._progress_nodeids_reported) == self._session.testscollected
+
     def pytest_runtest_logfinish(self, nodeid):
         assert self._session
         if self.verbosity <= 0 and self._show_progress_info:
@@ -465,15 +505,12 @@ class TerminalReporter:
             else:
                 progress_length = len(" [100%]")
 
-            main_color, _ = _get_main_color(self.stats)
-
             self._progress_nodeids_reported.add(nodeid)
-            is_last_item = (
-                len(self._progress_nodeids_reported) == self._session.testscollected
-            )
-            if is_last_item:
-                self._write_progress_information_filling_space(color=main_color)
+
+            if self._is_last_item:
+                self._write_progress_information_filling_space()
             else:
+                main_color, _ = self._get_main_color()
                 w = self._width_of_current_line
                 past_edge = w + progress_length + 1 >= self._screen_width
                 if past_edge:
@@ -497,9 +534,8 @@ class TerminalReporter:
                 )
             return " [100%]"
 
-    def _write_progress_information_filling_space(self, color=None):
-        if not color:
-            color, _ = _get_main_color(self.stats)
+    def _write_progress_information_filling_space(self):
+        color, _ = self._get_main_color()
         msg = self._get_progress_information_message()
         w = self._width_of_current_line
         fill = self._tw.fullwidth - w - 1
@@ -514,7 +550,7 @@ class TerminalReporter:
             # py < 1.6.0
             return self._tw.chars_on_current_line
 
-    def pytest_collection(self):
+    def pytest_collection(self) -> None:
         if self.isatty:
             if self.config.option.verbose >= 0:
                 self.write("collecting ... ", bold=True)
@@ -524,9 +560,9 @@ class TerminalReporter:
 
     def pytest_collectreport(self, report: CollectReport) -> None:
         if report.failed:
-            self.stats.setdefault("error", []).append(report)
+            self._add_stats("error", [report])
         elif report.skipped:
-            self.stats.setdefault("skipped", []).append(report)
+            self._add_stats("skipped", [report])
         items = [x for x in report.result if isinstance(x, pytest.Item)]
         self._numcollected += len(items)
         if self.isatty:
@@ -802,16 +838,32 @@ class TerminalReporter:
             for wr in warning_reports:
                 reports_grouped_by_message.setdefault(wr.message, []).append(wr)
 
-            title = "warnings summary (final)" if final else "warnings summary"
-            self.write_sep("=", title, yellow=True, bold=False)
-            for message, warning_reports in reports_grouped_by_message.items():
-                has_any_location = False
+            def collapsed_location_report(reports: List[WarningReport]):
+                locations = []
                 for w in warning_reports:
                     location = w.get_location(self.config)
                     if location:
-                        self._tw.line(str(location))
-                        has_any_location = True
-                if has_any_location:
+                        locations.append(location)
+
+                if len(locations) < 10:
+                    return "\n".join(map(str, locations))
+
+                counts_by_filename = collections.Counter(
+                    str(loc).split("::", 1)[0] for loc in locations
+                )
+                return "\n".join(
+                    "{0}: {1} test{2} with warning{2}".format(
+                        k, v, "s" if v > 1 else ""
+                    )
+                    for k, v in counts_by_filename.items()
+                )
+
+            title = "warnings summary (final)" if final else "warnings summary"
+            self.write_sep("=", title, yellow=True, bold=False)
+            for message, warning_reports in reports_grouped_by_message.items():
+                maybe_location = collapsed_location_report(warning_reports)
+                if maybe_location:
+                    self._tw.line(maybe_location)
                     lines = message.splitlines()
                     indented = "\n".join("  " + x for x in lines)
                     message = indented.rstrip()
@@ -909,7 +961,7 @@ class TerminalReporter:
             return
 
         session_duration = time.time() - self._sessionstarttime
-        (parts, main_color) = build_summary_stats_line(self.stats)
+        (parts, main_color) = self.build_summary_stats_line()
         line_parts = []
 
         display_sep = self.verbosity >= 0
@@ -949,7 +1001,7 @@ class TerminalReporter:
             failed = self.stats.get(stat, [])
             if not failed:
                 return
-            termwidth = self.writer.fullwidth
+            termwidth = self._tw.fullwidth
             config = self.config
             for rep in failed:
                 line = _get_line_with_reprcrash_message(config, rep, termwidth)
@@ -994,9 +1046,7 @@ class TerminalReporter:
             "x": show_xfailed,
             "X": show_xpassed,
             "f": partial(show_simple, "failed"),
-            "F": partial(show_simple, "failed"),
             "s": show_skipped,
-            "S": show_skipped,
             "p": partial(show_simple, "passed"),
             "E": partial(show_simple, "error"),
         }  # type: Mapping[str, Callable[[List[str]], None]]
@@ -1012,6 +1062,53 @@ class TerminalReporter:
             for line in lines:
                 self.write_line(line)
 
+    def _get_main_color(self) -> Tuple[str, List[str]]:
+        if self._main_color is None or self._known_types is None or self._is_last_item:
+            self._set_main_color()
+            assert self._main_color
+            assert self._known_types
+        return self._main_color, self._known_types
+
+    def _determine_main_color(self, unknown_type_seen: bool) -> str:
+        stats = self.stats
+        if "failed" in stats or "error" in stats:
+            main_color = "red"
+        elif "warnings" in stats or "xpassed" in stats or unknown_type_seen:
+            main_color = "yellow"
+        elif "passed" in stats or not self._is_last_item:
+            main_color = "green"
+        else:
+            main_color = "yellow"
+        return main_color
+
+    def _set_main_color(self) -> None:
+        unknown_types = []  # type: List[str]
+        for found_type in self.stats.keys():
+            if found_type:  # setup/teardown reports have an empty key, ignore them
+                if found_type not in KNOWN_TYPES and found_type not in unknown_types:
+                    unknown_types.append(found_type)
+        self._known_types = list(KNOWN_TYPES) + unknown_types
+        self._main_color = self._determine_main_color(bool(unknown_types))
+
+    def build_summary_stats_line(self) -> Tuple[List[Tuple[str, Dict[str, bool]]], str]:
+        main_color, known_types = self._get_main_color()
+
+        parts = []
+        for key in known_types:
+            reports = self.stats.get(key, None)
+            if reports:
+                count = sum(
+                    1 for rep in reports if getattr(rep, "count_towards_summary", True)
+                )
+                color = _color_for_type.get(key, _color_for_type_default)
+                markup = {color: True, "bold": color == main_color}
+                parts.append(("%d %s" % _make_plural(count, key), markup))
+
+        if not parts:
+            parts = [("no tests ran", {_color_for_type_default: True})]
+
+        return parts, main_color
+
 
 def _get_pos(config, rep):
     nodeid = config.cwd_relative_nodeid(rep.nodeid)
@@ -1100,50 +1197,6 @@ def _make_plural(count, noun):
     return count, noun + "s" if count != 1 else noun
 
 
-def _get_main_color(stats) -> Tuple[str, List[str]]:
-    known_types = (
-        "failed passed skipped deselected xfailed xpassed warnings error".split()
-    )
-    unknown_type_seen = False
-    for found_type in stats.keys():
-        if found_type not in known_types:
-            if found_type:  # setup/teardown reports have an empty key, ignore them
-                known_types.append(found_type)
-                unknown_type_seen = True
-
-    # main color
-    if "failed" in stats or "error" in stats:
-        main_color = "red"
-    elif "warnings" in stats or unknown_type_seen:
-        main_color = "yellow"
-    elif "passed" in stats:
-        main_color = "green"
-    else:
-        main_color = "yellow"
-
-    return main_color, known_types
-
-
-def build_summary_stats_line(stats):
-    main_color, known_types = _get_main_color(stats)
-
-    parts = []
-    for key in known_types:
-        reports = stats.get(key, None)
-        if reports:
-            count = sum(
-                1 for rep in reports if getattr(rep, "count_towards_summary", True)
-            )
-            color = _color_for_type.get(key, _color_for_type_default)
-            markup = {color: True, "bold": color == main_color}
-            parts.append(("%d %s" % _make_plural(count, key), markup))
-
-    if not parts:
-        parts = [("no tests ran", {_color_for_type_default: True})]
-
-    return parts, main_color
-
-
 def _plugin_nameversions(plugininfo) -> List[str]:
     values = []  # type: List[str]
     for plugin, dist in plugininfo:
